#!/bin/bash
#----------------------------------------------------------

# --- input/output dir
idir="/atmosdyn/schemms/oras5/raw/"
odir="/atmosdyn/schemms/oras5/"

tdir=$( mktemp -d )
cd $tdir



# --- create target grid
cat > targetgrid << endfile
gridtype=lonlat
xsize=1441
ysize=721
xfirst=-180.0
xinc=0.25
yfirst=-90.0
yinc=0.25
endfile

#----------------------------------------------------------
# Loop over all files
#----------------------------------------------------------

ln -s ${idir}/*T_01.nc .

for f in *T_01.nc
do

# ----- extract the date -------------
    # -- take final date of 5-day average
    date=${f:17:8}

    # ----- extract the base -------------
    base=${f:0:31}

    echo "Working on: " $base "Date: " $date
    
    # ----- prepare file names -----------
    ft=${base}"T_01.nc"
    
    # --- extract variables --------------
    cdo selvar,sosstsst  ${ft} sst.nc
    cdo selvar,sosaline  ${ft} sss.nc
    cdo selvar,sossheig  ${ft} ssh.nc
    cdo selvar,sowindsp  ${ft} w10.nc
    cdo selvar,soicecov  ${ft} ice.nc
    cdo selvar,sothedep  ${ft} the.nc
    cdo selvar,sohefldo  ${ft} hflx.nc
    cdo selvar,sowaflup  ${ft} wflx.nc
    cdo selvar,sosalflx  ${ft} salt.nc

    
    # --- combine them -------------------
    ncks -h -A ssh.nc sst.nc
    ncks -h -A sss.nc sst.nc
    ncks -h -A w10.nc sst.nc
    ncks -h -A ice.nc sst.nc
    ncks -h -A the.nc sst.nc
    ncks -h -A hflx.nc sst.nc
    ncks -h -A wflx.nc sst.nc
    ncks -h -A salt.nc sst.nc
    
    # --- set missing data
    cdo setmissval,-999.99 sst.nc tmp
    
    # ----- no netcdf4 support in Bergen
    #nccopy -k classic tmp tmp2
    #\mv tmp2 sst.nc
    #\rm tmp
    \mv tmp sst.nc
    
     # ----- set land to mdv
    ncap2 -O -h -s 'where(sosaline==0) sosstsst=-999.99' sst.nc sst.nc
    ncap2 -O -h -s 'where(sosaline==0) sossheig=-999.99' sst.nc sst.nc
    ncap2 -O -h -s 'where(sosaline==0) sowindsp=-999.99' sst.nc sst.nc
    ncap2 -O -h -s 'where(sosaline==0) soicecov=-999.99' sst.nc sst.nc
    ncap2 -O -h -s 'where(sosaline==0) sothedep=-999.99' sst.nc sst.nc
    ncap2 -O -h -s 'where(sosaline==0) sohefldo=-999.99' sst.nc sst.nc
    ncap2 -O -h -s 'where(sosaline==0) sowaflup=-999.99' sst.nc sst.nc
    ncap2 -O -h -s 'where(sosaline==0) sosalflx=-999.99' sst.nc sst.nc
    ncap2 -O -h -s 'where(sosaline==0) sosaline=-999.99' sst.nc sst.nc


    # ----- Interpolate to regular grid
    cdo remapbil,targetgrid sst.nc remaptmp.nc

   
    # ----- convert lat/lon/time to float
    ncrename -h -v time_counter,time remaptmp.nc
    ncrename -h -d time_counter,time remaptmp.nc

    ncap2 -O -h -s 'lat=float(lat)' remaptmp.nc remaptmp.nc
    ncap2 -O -h -s 'lon=float(lon)' remaptmp.nc remaptmp.nc
    ncap2 -O -h -s 'time=float(time)' remaptmp.nc remaptmp.nc
    
    # ----- set land to mdv
    ncap2 -O -h -s 'where(sosaline==0) sosstsst=-999.99' remaptmp.nc remaptmp.nc
    ncap2 -O -h -s 'where(sosaline==0) sossheig=-999.99' remaptmp.nc remaptmp.nc
    ncap2 -O -h -s 'where(sosaline==0) sowindsp=-999.99' remaptmp.nc remaptmp.nc
    ncap2 -O -h -s 'where(sosaline==0) soicecov=-999.99' remaptmp.nc remaptmp.nc
    ncap2 -O -h -s 'where(sosaline==0) sothedep=-999.99' remaptmp.nc remaptmp.nc
    ncap2 -O -h -s 'where(sosaline==0) sohefldo=-999.99' remaptmp.nc remaptmp.nc
    ncap2 -O -h -s 'where(sosaline==0) sowaflup=-999.99' remaptmp.nc remaptmp.nc
    ncap2 -O -h -s 'where(sosaline==0) sosalflx=-999.99' remaptmp.nc remaptmp.nc
    ncap2 -O -h -s 'where(sosaline==0) sosaline=-999.99' remaptmp.nc remaptmp.nc

    # ----- correct time axis
    yr=$(echo $date | cut -c1-4)
    mn=$(echo $date | cut -c5-6)
    dd=$(echo $date | cut -c7-8)
    hh="00"

    dat=${yr}'-'${mn}'-'${dd}
    time=${hh}':00'

    diff=$(( ($(date -u --date=${dat}' '${time} +%s) - $(date -u --date="1979-01-01 00:00" +%s) )/(60*60) ))

    # --- create new and updated time variable
    g=remaptmp.nc
    ncap2 -h -s "time[time]=${diff}.f" ${g} ${g} --ovr
    ncatted -h -O -a standard_name,time,o,c,"time" ${g}
    ncatted -h -O -a long_name,time,o,c,"time" ${g}
    ncatted -h -O -a units,time,o,c,"hours since 1979-01-01 00:00" ${g}
    ncatted -h -O -a calendar,time,o,c,"gregorian" ${g}

    # ---- set constants file
    ncatted -h -O -a constants_file_name,global,c,c,"S${date}_00" ${g}
    ncatted -h -O -a history,global,d,,'' ${g}
    ncatted -h -O -a CDI,global,d,,'' ${g}
    ncatted -h -O -a CDO,global,d,,'' ${g}


    # ----- move output ------------------
    echo "Created " "S"${date}"_00"
    \mv remaptmp.nc ${odir}/"S"${date}"_00"
    
    # ----- rename variables
    ncrename -h -v sosstsst,SST ${odir}/"S"${date}"_00"
    ncrename -h -v sosaline,SSS ${odir}/"S"${date}"_00"
    ncrename -h -v sossheig,HGT ${odir}/"S"${date}"_00"
    ncrename -h -v sowindsp,W10M ${odir}/"S"${date}"_00"
    ncrename -h -v soicecov,ICE ${odir}/"S"${date}"_00"
    ncrename -h -v sothedep,THCL ${odir}/"S"${date}"_00"
    ncrename -h -v sohefldo,HFLX ${odir}/"S"${date}"_00"
    ncrename -h -v sowaflup,WFLX ${odir}/"S"${date}"_00"
    ncrename -h -v sosalflx,SAFLX ${odir}/"S"${date}"_00"


    ncks -h -x -v time_counter_bnds ${odir}/"S"${date}"_00" tmp
    \mv tmp ${odir}/"S"${date}"_00"

    
    # ---- clean
    \rm sss.nc ssh.nc w10.nc ice.nc salt.nc hflx.nc sst.nc wflx.nc the.nc tmp

done


\rm targetgrid*


#----------------------------------------------------------
exit 0
#----------------------------------------------------------
