# ---------------------------------------------------------------------
# ---- Load libs
from   netCDF4 import Dataset
import numpy as np
# ---------------------------------------------------------------------

# ---------------------------------------------------------------------
# ---- Spherical Functions
# ---------------------------------------------------------------------


# --- Constants
REARTH = 6371001 # radius of the earth [m]
EPS = 1.0E-10
pi = np.pi

# --- Haversine
def haversine(rlon1,rlat1,rlon2,rlat2):
    """
    Returns the great circle (geodesic) distance measured in radians [0,pi)
    between positions (rlon1,rlat1) and (rlon2,rlat2) [rad] using the haversine
    formula.

        c = haversine(rlon1,rlat1,rlon2,rlat2)

    Reference: http://en.wikipedia.org/wiki/Haversine_formula

    2012-06-10 oyvind.breivik@ecmwf.int

    """

    dlat = rlat2 - rlat1  
    dlon = rlon2 - rlon1  
    a = np.sin(dlat/2)**2 + np.cos(rlat1)*np.cos(rlat2)*np.sin(dlon/2)**2  

    return 2*np.arcsin(np.sqrt(a))

# --- Angle to meridian
def spheredir(lon1,lat1,lon2,lat2):
    """
    Returns the local direction measured clockwise relative to north from
    point (lon1,lat1) towards (lon2,lat2) following a great circle
    (geodesic) path on a sphere.
   
        dr = spheredir(lon1,lat1,lon2,lat2)
 
    Arguments can be vectorized.
   
    Requires haversine, ang360.
   
    Reference: Abramowitz and Stegun (1970): Handbook of mathematical functions,
    Section 4.3.149.

    12 Aug 99, Oyvind.Breivik@nrsc.no.
    2012-06-10 Converted to Python 2.7, oyvind.breivik@ecmwf.int

    22 Oct 15  sebastian.schemm@uib.no:
    Assume a spherical triangle with shape
    
        B
       / \        where A,B,C denote the Angles
    a /   \ c     and a,b,c the great circle distances opposed to A,B,C
     /     \
    C ----- A
        b

    B is the North Pole (90,0), and we have Lat/lon of point at C and A
    We want the angle C

		        sin(A)   sin(B)   sin(C)
    The law of sine is:  ----  =  -----  =  -----
                        sin(a)   sin(b)   sin(c)

    The law of cosine is: cos(c) = cos(a)cos(b) + sin(a)sin(b)cos(C)

    We could solve the law of cosine for cos(C) and get 
    the great circle distances a,b,c. But, it is easier to obtain the correct
    quadrant using the atan2.

                    sin(C)  sin(c)*sin(B)/sin(b)
    Hence  tan(c) = ----- = --------------------
                    cos(C)  (cos(c)-cos(a)cos(b))/(sin(a)sin(b))

    which in some textbooks is reformulated even further.


    """

# --- get co-latitude (we stick to mathematical defintions)
    a = np.radians(90.0-lat1)
    c = np.radians(90.0-lat2)
    bb = np.radians(lon1-lon2)

# --- get great circle distance
    b = haversine(np.radians(lon1),np.radians(lat1),np.radians(lon2),np.radians(lat2))

# --- calculate nominator and denominator
    y = np.sin(c)*np.sin(bb)/(np.sin(b)+EPS)
    x = (np.cos(c)-np.cos(a)*np.cos(b))/(np.sin(a)*np.sin(b)+EPS)

# --- calculate artan(y/x)
    cc = np.arctan2(y,x)

# --- return and transform from -180:180 to 0:360
# Examples: If atan2 indicates -60, it is 60
#           If atan2 indicates +60, it is 300

    return ang360(360.0-np.degrees(cc))

# --- Maps an array of angles [deg] to [0,360).
def ang360(ang):
    """
    Maps an array of angles [deg] to [0,360).
    """

    return ang % 360
    

# --- 2D Rotation matrix
def rotmatrix(u,v,angle):
     """
     2D rotation matrix of vector (/u,v/) 
     by <angle> in counter-clockwise direction.

      /   cos(alpha) -sin(alpha)  \
      \   sin(alpha)  cos(alpha)  /


     22 Oct 15
     sebastian.schemm@uib.no

     Reference: http://mathworld.wolfram.com/RotationMatrix.html

     """

     a = u*np.cos(np.radians(angle)) - v*np.sin(np.radians(angle))
     b = u*np.sin(np.radians(angle)) + v*np.cos(np.radians(angle))

     return a,b
     
# ---------------------------------------------------------------------
# ---- Main
# ---------------------------------------------------------------------


# ---- Start the show
print ' Python ready to start'

# ---- Files to read in
ufile = 'INFILE_U'
vfile = 'INFILE_V'
tfile = 'INFILE_T'
wfile = 'INFILE_W'

# ---- Open the file
uid   = Dataset(ufile,'r')
vid   = Dataset(vfile,'r')
tid   = Dataset(tfile,'r')
wid   = Dataset(wfile,'r')


print ' Read in data'
# ---- Get the data
uin   = uid.variables ['vozocrtx'] [:]
vin   = vid.variables ['vomecrty'] [:]
tin   = tid.variables ['votemper'] [:]
sin   = tid.variables ['vosaline'] [:]
rin   = tid.variables ['vosigmat'] [:]
win   = wid.variables ['vovecrtz'] [:]
tim   = tid.variables ['time_counter'][:]

##### DEBUGGING #####
##### uin[0,:,:,:] = 10
##### vin[0,:,:,:] = 10

# --- read in grid at scalar point t to which U/V are staggered
orcalat  = tid.variables['nav_lat'][:]
orcalon  = tid.variables['nav_lon'][:]
orcavert = tid.variables['deptht'][:]

# --- Get the FillValue
mdv   = uid.variables['vozocrtx'].getncattr('_FillValue')

# --- Get the Time Unit
timunits = tid.variables['time_counter'].getncattr('units')

# ---------------------------------------------------------------------
# ---- De-staggering U/V/W Arakawa C Grid
#
# --- X-Y Axis            --- Z Axis
#          V                  Surface (meter=0)
#          |                    |
#     U----T----U         U---- T ---- U
#          |                    |
#          V                    W
#
# U is cyclic; V and W set to mdv at lower bottom
# ---------------------------------------------------------------------



# - numpy array of final size (all shapes are similar as t, wired for v and w :-)
ustag = np.zeros(tin.shape)
vstag = np.zeros(tin.shape)
tstag = np.zeros(tin.shape)
wstag = np.zeros(tin.shape)

tstag[:] = tin[:]

# - U in X (remember: Python starts at 0 as C++)
print ' De-staggering U'

for i in range(0,uin.shape[3]-2):
	ustag[:,:,:,i] = 0.5*(uin[:,:,:,i] + uin[:,:,:,i+1])

# - U at boundaries cyclic
	ustag[:,:,:,uin.shape[3]-1] = 0.5*(uin[:,:,:,uin.shape[3]-1] + uin[:,:,:,0])

# - V in Y
print ' De-staggering V'

for j in range(0,vin.shape[2]-2):
	vstag[:,:,j,:] = 0.5*(vin[:,:,j,:] + vin[:,:,j+1,:])


# - V at boundaries missing
	vstag[:,:,vin.shape[2]-1,:] = mdv


# - W in Z (1 w level is)
print ' De-staggering W'

for k in range(0,win.shape[1]-2):
    wstag[:,k,:,:] = 0.5*(win[:,k,:,:] + win[:,k+1,:,:])

# - W at bottom missing
    wstag[:,win.shape[1]-1,:,:] = mdv

# ---------------------------------------------------------------------
# ---- Transform U/V from ORCA grid to regular lat/lon grid  
# ---------------------------------------------------------------------

print ' Get angle for rotation of U/V'

# --- array of angle
angle   = np.zeros(orcalat.shape)

# --- shape
nx = ustag.shape[3]
ny = ustag.shape[2]
nz = ustag.shape[1]
nt = ustag.shape[0]

# ---------------------------------------------------------------------
# --- Identify angle (clockwise) between north pol (meridian) and orca meridian
# ---------------------------------------------------------------------



# ---- Maybe this can be coded more efficient?

#for i in range(0,nx-1):
#	for j in range(0,ny-1):
# - northern boundary
#		if j==(ny-1):
#			angle[j,i] = spheredir(orcalon[j-1,i],orcalat[j-1,i],orcalon[j,i],orcalat[j,i])
# - interior
#		if j!=(ny-1):
#			angle[j,i] = spheredir(orcalon[j,i],orcalat[j,i],orcalon[j+1,i],orcalat[j+1,i])

# ---- Use angle to rotate U/V
#print ' Rotation of U/V '
#print '   (This takes horribly long - how to improve?)'

# - numpy array of final size (similar as t)
#u = np.zeros(tin.shape)
#v = np.zeros(tin.shape)

# - rotmatrix rotates counterclockwise
#u, v  = rotmatrix(ustag,vstag,angle)

u = ustag
v = vstag

# ---------------------------------------------------------------------
# Create a land/sea mask using salinity
# ---------------------------------------------------------------------
#land = np.zeros(tin.shape)

#land[sin<=1e-10] = 0
#land[sin>1e-10]  = 1

     
# ---------------------------------------------------------------------
# ---- Write out U/V/T on one file  
# ---------------------------------------------------------------------

print ' Write out'

ncout = Dataset('tmp.nc', 'w', format='NETCDF3_64BIT')

# -- define dimensions
ncout.createDimension('time', size=None)
ncout.createDimension('y', ny)
ncout.createDimension('x', nx)
ncout.createDimension('lev', nz)
#ncout.createDimension('tbnds', 2)


# -- create coordinate variables
times = ncout.createVariable('time', 'f', ('time'))
lon   = ncout.createVariable('lon2d', 'f', ('y','x'))
lat   = ncout.createVariable('lat2d', 'f', ('y','x'))
vert  = ncout.createVariable('lev', 'f', ('lev'))


# -- attributes
lat.units     = 'degrees_north'
lon.units     = 'degrees_east'

lat.long_name = 'Latitude'
lon.long_name = 'Longitude'

lat.standard_name = 'latitude'
lon.standard_name = 'longitude'

lon.valid_min = -180.
lon.valid_max = 180.

lat.valid_min = -76.98
lat.valid_max = 89.93

vert.units     = 'm'
vert.axis      = 'Z'
vert.positive  = 'down'
vert.valid_min = 0.51
vert.valid_max = 5902.058
vert.long_name = 'Vertical V levels'

times.standard_name = 'time'
times.calendar      = 'gregorian'
times.axis          = 't'
times.long_name     = 'Time'
times.units         = str(timunits)
times.avg           = '5-day average'

# write data to coordinate vars.
lat[:]   = orcalat
lon[:]   = orcalon
vert[:]  = orcavert
# add five days to start
times[:] = 345600

# create variables 
ncu = ncout.createVariable('U','f',('time','lev','y','x'),fill_value=mdv)
ncv = ncout.createVariable('V','f',('time','lev','y','x'),fill_value=mdv)
nct = ncout.createVariable('T','f',('time','lev','y','x'),fill_value=mdv)
ncs = ncout.createVariable('S','f',('time','lev','y','x'),fill_value=mdv)
ncw = ncout.createVariable('W','f',('time','lev','y','x'),fill_value=mdv)
ncr = ncout.createVariable('SIGMAT','f',('time','lev','y','x'),fill_value=mdv)

#ncl = ncout.createVariable('mask','f4',('time','lev','y','x'),fill_value=mdv)

##### DEBUGGING #####
##### ncd = ncout.createVariable('diffU','f4',('time','lev','y','x'),fill_value=mdv)


# set the units attribute.
ncu.units = 'm/s'
ncv.units = 'm/s'
ncw.units = 'm/s'
nct.units = 'C'
ncs.units = 'PSU'
ncr.units = ''

ncv.standard_name = 'Meridional Current'
ncu.standard_name = 'Zonal Current'
nct.standard_name = 'Potential Temperature'
ncs.standard_name = 'Salinity'
ncw.standard_name = 'Vertical Velocity'
ncr.standard_name = 'Sigma-theta'

#ncl.standard_name = 'land ocean mask'

ncv.coordinates = 'time lev lat2d lon2d'
ncu.coordinates = 'time lev lat2d lon2d'
nct.coordinates = 'time lev lat2d lon2d'
ncs.coordinates = 'time lev lat2d lon2d'
ncr.coordinates = 'time lev lat2d lon2d'
ncw.coordinates = 'time lev lat2d lon2d'
##### DEBUGGING #####
#####ncd.coordinates = 'time lev lat2d lon2d'

# transform sigmat from NEMO convention
rin = (rin*1035)+1035-1000

# write data to variables.
ncu[:] = u
ncv[:] = v
ncw[:] = wstag
ncs[:] = sin
nct[:] = tstag
ncr[:] = rin
#ncl[:] = land


##### DEBUGGING #####
##### ncd[:] = u-uin

# close the file.
ncout.close()

